import dataclasses
import uuid
from datetime import timedelta
from pstats import SortKey

import pytest
from dataclass_dict_convert.convert import TypeConvertorError
from imecilabt_utils.urn_util import UrnError

from tests.test_usage_statistics import TEST_GPU_OVERVIEW_OBJ, \
    TEST_GPU_OVERVIEW_JSON, TEST_GPULAB_USAGE_STATISTICS_JSON, TEST_GPULAB_USAGE_STATISTICS_OBJ

from imecilabt.gpulab.model.job2 import *

from dataclass_dict_convert import parse_rfc3339


def _parse_date(date: str) -> datetime.datetime:
    return parse_rfc3339(date)


TEST_JOB2_REQUEST_RESOURCES_JSON = """{
  "cpus": 4,
  "gpus": 2,
  "cpuMemoryGb": 8,

  "minCudaVersion": 10,
  "clusterId": 7,
  "gpuModel": ["1080", "v100"],
  "slaveName": "slave4B",
  "slaveInstanceId": "inst10",
  "features": ["SHARED_PROJECT_STORAGE", "FAST_SCRATCH_PROJECT_STORAGE", 
               "PUBLIC_IPV4", "PUBLIC_IPV6", "UNFIREWALLED_PORTS", "SSH_ACCESS"]
}"""
TEST_JOB2_REQUEST_SCHEDULING_JSON = """{
  "interactive": false,
  "minDuration": "5 hour",
  "restartable": false,
  "reservationId": "123456-789a-bcde-f012-34567890",

  "maxDuration": "5 hour",
  "maxSimultaneousJobs": {
    "bucketName": "my-3jobs-demo",
    "bucketMax": 3
  },
  "notBefore": "2020-03-17T05:00:13Z",
  "notAfter": "2020-03-17T18:00:13Z"
}"""

# minDuration was named killableAfter for a while!
TEST_JOB2_REQUEST_SCHEDULING_JSON_BACKWARD_COMP = """{
  "interactive": false,
  "killableAfter": "5 hour",
  "restartable": false,
  "reservationId": "123456-789a-bcde-f012-34567890",

  "maxDuration": "5 hour",
  "maxSimultaneousJobs": {
    "bucketName": "my-3jobs-demo",
    "bucketMax": 3
  },
  "notBefore": "2020-03-17T05:00:13Z",
  "notAfter": "2020-03-17T18:00:13Z"
}"""

TEST_JOB2_REQUEST_DOCKER_JSON = """{
    "image": "gitlab.ilabt.imec.be:4567/ilabt/gpu-docker-stacks/tensorflow-notebook:83755206e3198b4914852bd640c102cb757fc02e",
    "command": [ "/project/start-demo.sh", "demo" ],
    "environment": {
      "MYDEMOENVVAR": "demo"
    },
    "storage": [
      {
        "containerPath": "/proj",
        "hostPath": "/project"
      },
      {
        "hostPath": "/project_scratch"
      },
      {
        "containerPath": "/project_scratch2"
      },
      {
        "containerPath": "/foobar",
        "hostPath": "tmpfs",
        "sizeGb": 5
      }
    ],
    "portMappings": [
      {
        "containerPort": 8888,
        "hostPort": 80
      }
    ],

    "projectGidVariableName": "NB_GID",
    "user": "root",
    "groupAdd": ["demogroup"],
    "workingDir": "/root/"
}"""
TEST_JOB2_REQUEST_EXTRA_JSON = """{
   "sshPubKeys": ["ssh-rsa test1"],
   "emailOnEnd": ["end@example.com"],
   "emailOnQueue": ["queue1@example.com", "queue2@example.com"],
   "emailOnRun": ["run@example.com"],
   "emailOnHalt": ["halted@example.com"],
   "emailOnRestart": ["restart@example.com"]
}"""

TEST_JOB2_REQUEST_JSON = """{
    "resources": """+TEST_JOB2_REQUEST_RESOURCES_JSON+""",
    "scheduling": """+TEST_JOB2_REQUEST_SCHEDULING_JSON+""",
    "docker": """+TEST_JOB2_REQUEST_DOCKER_JSON+""",
    "extra": """+TEST_JOB2_REQUEST_EXTRA_JSON+"""
}"""

TEST_JOB2_REQUEST_JSON_BACKWARD_COMP = """{
    "resources": """+TEST_JOB2_REQUEST_RESOURCES_JSON+""",
    "scheduling": """+TEST_JOB2_REQUEST_SCHEDULING_JSON_BACKWARD_COMP+""",
    "docker": """+TEST_JOB2_REQUEST_DOCKER_JSON+""",
    "extra": """+TEST_JOB2_REQUEST_EXTRA_JSON+"""
}"""

TEST_JOB2_REQUEST_RESOURCES_OBJ_A = JobRequestResources(
    cpus=4,
    gpus=2,
    cpu_memory_gb=8,
    gpu_memory_gb=None,
    min_cuda_version=10,
    cluster_id=7,
    gpu_model=["1080", "v100"],
    slave_name="slave4B",
    slave_instance_id="inst10",
    features=[
        "SHARED_PROJECT_STORAGE",
        "FAST_SCRATCH_PROJECT_STORAGE",
        "PUBLIC_IPV4",
        "PUBLIC_IPV6",
        "UNFIREWALLED_PORTS",
        "SSH_ACCESS",
    ],
)
TEST_JOB2_REQUEST_RESOURCES_OBJ_B = JobRequestResources(
    cpus=1,
    gpus=0,
    cpu_memory_gb=4,
    gpu_memory_gb=4,
)

TEST_JOB2_REQUEST_SCHEDULING_OBJ_A = JobRequestScheduling(
    interactive=False,
    min_duration="5 hour",
    restartable=False,
    reservation_ids=["123456-789a-bcde-f012-34567890"],
    max_duration="5 hour",
    max_simultaneous_jobs=MaxSimultaneousJobs(bucket_max=3, bucket_name="my-3jobs-demo"),
    not_before=_parse_date("2020-03-17T05:00:13Z"),
    not_after=_parse_date("2020-03-17T18:00:13Z"),
)

TEST_JOB2_REQUEST_DOCKER_OBJ_A = JobRequestDocker(
    image="gitlab.ilabt.imec.be:4567/ilabt/gpu-docker-stacks/tensorflow-notebook:83755206e3198b4914852bd640c102cb757fc02e",
    command=["/project/start-demo.sh", "demo"],
    environment={"MYDEMOENVVAR": "demo"},
    project_gid_variable_name="NB_GID",
    user="root",
    group_add=["demogroup"],
    working_dir="/root/",
    storage=[
        JobStorage(container_path="/proj", host_path="/project", size_gb=None),
        JobStorage(container_path="/project_scratch", host_path="/project_scratch", size_gb=None),
        JobStorage(container_path="/project_scratch2", host_path="/project_scratch2", size_gb=None),
        JobStorage(container_path="/foobar", host_path="tmpfs", size_gb=5),
    ],
    port_mappings=[JobPortMapping(container_port=8888, host_port=80)],
)
TEST_JOB2_REQUEST_DOCKER_OBJ_B = JobRequestDocker(
    image="debian:stable"
)

TEST_JOB2_REQUEST_EXTRA_OBJ_A = JobRequestExtra(
    ssh_pub_keys=["ssh-rsa test1"],
    email_on_end=["end@example.com"],
    email_on_queue=["queue1@example.com", "queue2@example.com"],
    email_on_run=["run@example.com"],
    email_on_halt=["halted@example.com"],
    email_on_restart=["restart@example.com"],
)

TEST_JOB2_REQUEST_OBJ_A = JobRequest(
    resources=TEST_JOB2_REQUEST_RESOURCES_OBJ_A,
    scheduling=TEST_JOB2_REQUEST_SCHEDULING_OBJ_A,
    docker=TEST_JOB2_REQUEST_DOCKER_OBJ_A,
    extra=TEST_JOB2_REQUEST_EXTRA_OBJ_A,
)
TEST_JOB2_REQUEST_OBJ_B = JobRequest(
    resources=TEST_JOB2_REQUEST_RESOURCES_OBJ_B,
    docker=TEST_JOB2_REQUEST_DOCKER_OBJ_B,
)



def test_job2_request_from_json():
    out_job2_request: JobRequest = JobRequest.from_json(TEST_JOB2_REQUEST_JSON)
    assert out_job2_request == TEST_JOB2_REQUEST_OBJ_A


def test_job2_request_from_json_backward_comp():
    out_job2_request: JobRequest = JobRequest.from_json(TEST_JOB2_REQUEST_JSON_BACKWARD_COMP)
    assert out_job2_request == TEST_JOB2_REQUEST_OBJ_A


def test_job2_request_docker_from_json1():
    in_json = """{
            "image": "dummy",
            "command": "test direct string",
            "groupAdd": "demogroup"
        }
    """
    out_job2_request_docker: JobRequestDocker = JobRequestDocker.from_json(in_json)

    assert out_job2_request_docker == JobRequestDocker(
        image="dummy",
        command="test direct string",
        environment={},
        project_gid_variable_name=None,
        user=None,
        group_add=["demogroup"],
        working_dir=None,
        storage=[],
        port_mappings=[],
    )


def test_job2_request_docker_from_json2():
    in_json = """{
            "image": "dummy",
            "storage": [
              "/project_scratch",
              {
                "containerPath": "/proj",
                "hostPath": "/project"
              },
              "/project_scratch2",
              {
                "containerPath": "/baz",
                "hostPath": "tmpfs",
                "sizeGb": 4
              }
            ],
            "portMappings": [
              77,
              {
                "containerPort": 8888,
                "hostPort": 80
              }
            ]
        }
    """
    out_job2_request_docker: JobRequestDocker = JobRequestDocker.from_json(in_json)

    assert out_job2_request_docker == JobRequestDocker(
        image="dummy",
        command=[],
        environment={},
        project_gid_variable_name=None,
        user=None,
        group_add=[],
        working_dir=None,
        storage=[
            JobStorage(container_path="/project_scratch", host_path="/project_scratch", size_gb=None),
            JobStorage(container_path="/proj", host_path="/project", size_gb=None),
            JobStorage(container_path="/project_scratch2", host_path="/project_scratch2", size_gb=None),
            JobStorage(container_path="/baz", host_path="tmpfs", size_gb=4),
        ],
        port_mappings=[
            JobPortMapping(container_port=77, host_port=None),
            JobPortMapping(container_port=8888, host_port=80),
        ],
    )


TEST_JOB2_STATE_RESOURCES_JSON = """{
   "clusterId": 7,
   "cpuIds": [4, 5],
   "gpuIds": [1],
   "cpuMemoryGb": 8,
   "gpuMemoryGb": 4,
 
   "slaveHost": "hostname.example.com",
   "slaveName": "hostname",
   "slaveInstanceId": "inst8",
   "slaveInstancePid": 5097,
   "workerId": 11,
 
   "sshHost": "host.example.com",
   "sshPort": 22,
   "sshUsername": "ABCDEF",
   "sshProxyHost": "bastion.example.com",
   "sshProxyPort": 2222,
   "sshProxyUsername": "fffdemo",
 
   "gpuDetails": """+TEST_GPU_OVERVIEW_JSON+""",
   "portMappings": [
     {
       "containerPort": 8888,
       "hostIp": "0.0.0.0",
       "hostPort": 32935
     }
   ]
 }"""

TEST_JOB2_STATE_EVENTTIMES_JSON = """{
   "created": "2020-03-17T09:59:19Z",
   "statusUpdated": "2020-03-17T10:00:13Z",

   "QUEUED": "2020-03-17T09:59:20Z",
   "ASSIGNED": "2020-03-17T08:59:24Z",
   "STARTING": "2020-03-17T09:59:21Z",
   "RUNNING": "2020-03-17T09:59:22Z",
   "FINISHED": "2020-03-17T09:59:23Z",
   "FAILED": "2020-03-17T09:59:24Z",
   "CANCELLED": "2020-03-17T09:59:25Z",
   "DELETED": "2020-03-17T09:59:26Z",
   "longRunNotify": "2020-03-18T09:59:21Z",
    "MUSTHALT": "2020-03-17T09:59:27Z",
    "HALTING": "2020-03-17T09:59:28Z",
    "HALTED": "2020-03-17T09:59:29Z"
 }"""

TEST_JOB2_STATE_SCHEDULING_JSON = """{
   "assignedClusterId": 7,
   "assignedInstanceId": "inst8",
   "assignedSlaveName": "hostname",
   "queuedExplanations": [],
   "withinMaxSimultaneousJobs": true,
   "tallyIncrement": 123.4
}"""

TEST_JOB2_STATE_JSON = """{
     "status": "RUNNING",
     "resources": """+TEST_JOB2_STATE_RESOURCES_JSON+""",
     "eventTimes": """+TEST_JOB2_STATE_EVENTTIMES_JSON+""",
     "scheduling": """+TEST_JOB2_STATE_SCHEDULING_JSON+""",
     "finalUsageStatistics": """+TEST_GPULAB_USAGE_STATISTICS_JSON+"""
}"""

TEST_JOB2_STATE_RESOURCES_OBJ_A = JobStateResources(
    cluster_id=7,
    cpu_ids=[4, 5],
    gpu_ids=[1],
    cpu_memory_gb=8,
    gpu_memory_gb=4,
    slave_host="hostname.example.com",
    slave_name="hostname",
    slave_instance_id="inst8",
    slave_instance_pid=5097,
    worker_id=11,
    ssh_host="host.example.com",
    ssh_port=22,
    ssh_username="ABCDEF",
    ssh_proxy_host="bastion.example.com",
    ssh_proxy_port=2222,
    ssh_proxy_username="fffdemo",
    port_mappings=[JobPortMapping(container_port=8888, host_ip="0.0.0.0", host_port=32935)],
    gpu_details=TEST_GPU_OVERVIEW_OBJ,
)

TEST_JOB2_STATE_SCHEDULING_OBJ_A = JobStateScheduling(
    assigned_cluster_id=7,
    assigned_instance_id="inst8",
    assigned_slave_name="hostname",
    queued_explanations=[],
    within_max_simultaneous_jobs=True,
    tally_increment=123.4,
)

TEST_JOB2_STATE_SCHEDULING_OBJ_B = JobStateScheduling(
    queued_explanations=["Not enough free resources too run this Job"]
)

TEST_JOB2_STATE_EVENTTIMES_OBJ_A = JobEventTimes(
    created=_parse_date("2020-03-17T09:59:19Z"),
    status_updated=_parse_date("2020-03-17T10:00:13Z"),
    QUEUED=_parse_date("2020-03-17T09:59:20Z"),
    ASSIGNED=_parse_date("2020-03-17T08:59:24Z"),
    STARTING=_parse_date("2020-03-17T09:59:21Z"),
    RUNNING=_parse_date("2020-03-17T09:59:22Z"),
    FINISHED=_parse_date("2020-03-17T09:59:23Z"),
    FAILED=_parse_date("2020-03-17T09:59:24Z"),
    CANCELLED=_parse_date("2020-03-17T09:59:25Z"),
    DELETED=_parse_date("2020-03-17T09:59:26Z"),
    MUSTHALT=_parse_date("2020-03-17T09:59:27Z"),
    HALTING=_parse_date("2020-03-17T09:59:28Z"),
    HALTED=_parse_date("2020-03-17T09:59:29Z"),
    long_run_notify=_parse_date("2020-03-18T09:59:21Z"),
)
TEST_JOB2_STATE_EVENTTIMES_OBJ_B = JobEventTimes(
    created=_parse_date("2020-04-15T09:59:19Z"),
    status_updated=_parse_date("2020-04-15T12:34:56Z"),
    QUEUED=_parse_date("2020-04-15T09:59:20Z"),
)
TEST_JOB2_STATE_OBJ_A = JobState(
    status=JobStatus.RUNNING,
    resources=TEST_JOB2_STATE_RESOURCES_OBJ_A,
    scheduling=TEST_JOB2_STATE_SCHEDULING_OBJ_A,
    event_times=TEST_JOB2_STATE_EVENTTIMES_OBJ_A,
    final_usage_statistics=TEST_GPULAB_USAGE_STATISTICS_OBJ,
)

TEST_JOB2_STATE_OBJ_B = JobState(
    status=JobStatus.QUEUED,
    resources=None,
    scheduling=TEST_JOB2_STATE_SCHEDULING_OBJ_B,
    event_times=TEST_JOB2_STATE_EVENTTIMES_OBJ_B,
    final_usage_statistics=None,
)


def test_job2_state_from_json():
    out_job2_state: JobState = JobState.from_json(TEST_JOB2_STATE_JSON)
    assert out_job2_state == TEST_JOB2_STATE_OBJ_A


TEST_JOB2_OWNER_JSON = """{
    "userUrn": "urn:publicid:IDN+example.com+user+tester",
    "userEmail": "tester@example.com",
    "projectUrn": "urn:publicid:IDN+example.com+project+testproj"
}"""

TEST_JOB2_OWNER_OBJ_A = JobOwner(
    user_urn="urn:publicid:IDN+example.com+user+tester",
    user_email="tester@example.com",
    project_urn="urn:publicid:IDN+example.com+project+testproj",
)

TEST_JOB2_OWNER_OBJ_B = JobOwner(
    user_urn="urn:publicid:IDN+example.com+user+testerB",
    user_email="testerB@example.com",
    project_urn="urn:publicid:IDN+example.com+project+testprojB",
)

def test_job_owner1():
    expected = TEST_JOB2_OWNER_OBJ_A
    actual = JobOwner.from_json(TEST_JOB2_OWNER_JSON)
    assert actual == expected

TEST_JOB2_UUID_A   = str(uuid.uuid4())
TEST_JOB2_UUID_B   = "0ae7f528-ac52-4362-8d2f-ab6756977a1c"
TEST_JOB2_UUID_NEW = str(uuid.uuid4())
TEST_JOB2_UUID_B_2 = "0a49e527-8e86-4803-9162-ed6509206baf"
TEST_JOB2_UUID_B_5 = "0ae7f61f-c617-46a9-bbb0-060041ab537c"


TEST_JOB2_JSON = """{
    "uuid": """+'"'+TEST_JOB2_UUID_A+'"'+""",
    "name": "testJobA",
    "deploymentEnvironment": "testing",
    "request": """+TEST_JOB2_REQUEST_JSON+""",
    "description": "Test Job A",
    "owner": """+TEST_JOB2_OWNER_JSON+""",
    "state": """+TEST_JOB2_STATE_JSON+"""
}"""

TEST_JOB2_JSON_BACKWARD_COMP = """{
    "uuid": """+'"'+TEST_JOB2_UUID_A+'"'+""",
    "name": "testJobA",
    "deploymentEnvironment": "testing",
    "request": """+TEST_JOB2_REQUEST_JSON_BACKWARD_COMP+""",
    "description": "Test Job A",
    "owner": """+TEST_JOB2_OWNER_JSON+""",
    "state": """+TEST_JOB2_STATE_JSON+"""
}"""

TEST_JOB2_OBJ_A = Job(
    uuid=TEST_JOB2_UUID_A,
    name="testJobA",
    deployment_environment="testing",
    request=TEST_JOB2_REQUEST_OBJ_A,
    description="Test Job A",
    owner=TEST_JOB2_OWNER_OBJ_A,
    state=TEST_JOB2_STATE_OBJ_A,
)


TEST_JOB2_OBJ_B = Job(
    uuid=TEST_JOB2_UUID_B,
    name="TestJobB",
    deployment_environment="testing",
    request=TEST_JOB2_REQUEST_OBJ_B,
    description="Test Job B",
    owner=TEST_JOB2_OWNER_OBJ_B,
    state=TEST_JOB2_STATE_OBJ_B,
)

TEST_JOB2_JSON_B = TEST_JOB2_OBJ_B.to_json()

TEST_JOB2_OBJ_C: Job = dataclasses.replace(
    TEST_JOB2_OBJ_A,
    uuid='7287c5dd-715e-487e-887d-2a2c8e52651a',
    name='TestJobC',
    owner=dataclasses.replace(
        TEST_JOB2_OBJ_A.owner,
        user_urn="urn:publicid:IDN+example.com+user+testerC"
    ),
    state=dataclasses.replace(
        TEST_JOB2_STATE_OBJ_A,
        resources=dataclasses.replace(
            TEST_JOB2_STATE_RESOURCES_OBJ_A,
            slave_name='hostnameB',
            slave_instance_id='inst8',
        ),
    )
)

TEST_JOB2_JSON_C = TEST_JOB2_OBJ_C.to_json()

def test_job_a():
    expected = TEST_JOB2_OBJ_A
    actual = Job.from_json(TEST_JOB2_JSON)
    assert actual == expected


def test_job_a_backward_comp():
    expected = TEST_JOB2_OBJ_A
    actual = Job.from_json(TEST_JOB2_JSON_BACKWARD_COMP)
    assert actual == expected


def test_job_a_to_from_json():
    expected = TEST_JOB2_OBJ_A
    actual = Job.from_json(TEST_JOB2_OBJ_A.to_json())
    assert actual == expected


def test_job_a_to_from_dict():
    expected = TEST_JOB2_OBJ_A
    actual = Job.from_dict(TEST_JOB2_OBJ_A.to_dict())
    assert actual == expected


def test_jobs_from_dict_1():
    input = []
    expected = []
    actual: List[Job] = Job.from_dict_list(input)
    assert actual == expected


def test_jobs_from_dict_2():
    input = [TEST_JOB2_OBJ_A.to_dict()]
    expected = [TEST_JOB2_OBJ_A]
    # should work, but fails de decode RFC3339 date correctly:
    # actual: List[Job] = Job.schema().load(input, many=True)
    actual: List[Job] = Job.from_dict_list(input)
    assert actual == expected


def test_job_a_to_dict_eventtimes():
    actual_dict = TEST_JOB2_OBJ_A.to_dict()
    assert 'QUEUED' in actual_dict['state']['eventTimes'].keys()
    assert actual_dict['state']['eventTimes']['QUEUED'] == '2020-03-17T09:59:20Z'


def test_job_from_json_valid_owner_check1():
    in_json = '''{
       "uuid": "'''+TEST_JOB2_UUID_A+'''",
       "name": "check",
       "deploymentEnvironment": "staging",
       "request": {
          "docker": { "command":"nvidia-smi", "image":"nvidia/cuda:10.1-cudnn7-devel-ubuntu18.04"},
          "resources": { "clusterId":1,"cpus":1,"gpus":1,"cpuMemoryGb":2 }
       },
       "owner": {
          "projectUrn": "urn:publicid:IDN+example.com+project+good",
          "userUrn": "urn:publicid:IDN+example.com+user+good",
          "userEmail": "good@example.com"
       }
    }'''
    actual = Job.from_json(in_json)
    assert actual.owner.project_urn == "urn:publicid:IDN+example.com+project+good"
    assert actual.owner.user_urn == "urn:publicid:IDN+example.com+user+good"
    assert actual.owner.user_email == "good@example.com"


def test_job_from_json_valid_owner_checkfail1():
    in_json = '''{
       "uuid": "'''+TEST_JOB2_UUID_A+'''",
       "name": "check",
       "deploymentEnvironment": "staging",
       "request": {
          "docker": { "command":"nvidia-smi", "image":"nvidia/cuda:10.1-cudnn7-devel-ubuntu18.04"},
          "resources": { "clusterId":1,"cpus":1,"gpus":1,"cpuMemoryGb":2 }
       },
       "owner": {
          "projectUrn": "bad",
          "userUrn": "urn:publicid:IDN+example.com+user+good",
          "userEmail": "good@example.com"
       }
    }'''
    # with pytest.raises(UrnError):
    with pytest.raises(TypeConvertorError):
        actual = Job.from_json(in_json)


def test_job_from_json_valid_owner_checkfail2():
    in_json = '''{
       "uuid": "'''+TEST_JOB2_UUID_A+'''",
       "name": "check",
       "deploymentEnvironment": "staging",
       "request": {
          "docker": { "command":"nvidia-smi", "image":"nvidia/cuda:10.1-cudnn7-devel-ubuntu18.04"},
          "resources": { "clusterId":1,"cpus":1,"gpus":1,"cpuMemoryGb":2 }
       },
       "owner": {
          "projectUrn": "urn:publicid:IDN+example.com+project+good",
          "userUrn": "bad",
          "userEmail": "good@example.com"
       }
    }'''
    # with pytest.raises(UrnError):
    with pytest.raises(TypeConvertorError):
        actual = Job.from_json(in_json)


def test_job_from_json_valid_owner_checkfail3():
    in_json = '''{
       "uuid": "'''+TEST_JOB2_UUID_A+'''",
       "name": "check",
       "deploymentEnvironment": "staging",
       "request": {
          "docker": { "command":"nvidia-smi", "image":"nvidia/cuda:10.1-cudnn7-devel-ubuntu18.04"},
          "resources": { "clusterId":1,"cpus":1,"gpus":1,"cpuMemoryGb":2 }
       },
       "owner": {
          "projectUrn": "urn:publicid:IDN+example.com+project+good",
          "userUrn": "urn:publicid:IDN+example.com+user+good",
          "userEmail": "bad"
       }
    }'''
    # with pytest.raises(ValueError):
    with pytest.raises(TypeConvertorError):
        actual = Job.from_json(in_json)


def test_job_from_json_valid_owner_checkfail4():
    in_json = '''{
       "uuid": "'''+TEST_JOB2_UUID_A+'''",
       "name": "check",
       "deploymentEnvironment": "staging",
       "request": {
          "docker": { "command":"nvidia-smi", "image":"nvidia/cuda:10.1-cudnn7-devel-ubuntu18.04"},
          "resources": { "clusterId":1,"cpus":1,"gpus":1,"cpuMemoryGb":2 }
       },
       "owner": {
          "projectUrn": "urn:publicid:IDN+example.com+project+good",
          "userUrn": "urn:publicid:IDN+example.com+bad+good",
          "userEmail": "good@example.com"
       }
    }'''
    # with pytest.raises(UrnError):
    with pytest.raises(TypeConvertorError):
        actual = Job.from_json(in_json)


def test_job_from_json_valid_owner_checkfail5():
    in_json = '''{
       "uuid": "'''+TEST_JOB2_UUID_A+'''",
       "name": "check",
       "deploymentEnvironment": "staging",
       "request": {
          "docker": { "command":"nvidia-smi", "image":"nvidia/cuda:10.1-cudnn7-devel-ubuntu18.04"},
          "resources": { "clusterId":1,"cpus":1,"gpus":1,"cpuMemoryGb":2 }
       },
       "owner": {
          "projectUrn": "urn:publicid:IDN+example.com+bad+good",
          "userUrn": "urn:publicid:IDN+example.com+user+good",
          "userEmail": "good@example.com"
       }
    }'''
    # with pytest.raises(UrnError):
    with pytest.raises(TypeConvertorError):
        actual = Job.from_json(in_json)


def test_job_from_json_no_uuid():
    in_json = '''{
       "name":"NVIDIA SMI",
       "description":"Writes the output of the command 'nvidia-smi' to the log and exits",
       "request": {
          "docker": {
             "command":"nvidia-smi",
             "image":"nvidia/cuda:10.1-cudnn7-devel-ubuntu18.04",
             "storage":[],
             "portMappings":[],
             "groupAdd":[]
          },
          "resources": {
             "clusterId":1,"cpus":1,"gpus":1,"cpuMemoryGb":2,"gpuMemoryGb":4,"minCudaVersion":10,"features":[],"gpuModel":[]
          },
          "scheduling": {
             "interactive":false,"restartable":false
          },
          "extra": {
             "sshPubKeys":[],"emailOnQueue":[],"emailOnRun":[],"emailOnEnd":[],"emailOnHalt":[],"emailOnRestart":[]
          }
       },
       "owner": {
          "projectUrn": "urn:publicid:IDN+example.com+project+good",
          "userUrn": "urn:publicid:IDN+example.com+user+good",
          "userEmail": "good@example.com"
       },
       "deploymentEnvironment": "staging"
    }'''
    actual = Job.from_json(in_json)
    assert actual.uuid is None
    assert actual.name == "NVIDIA SMI"


def test_port_mapping_from_docker_dict_a():
    d = {"8888/tcp": [{"HostIp": "0.0.0.0", "HostPort": "32769"}]}
    container_port = "8888/tcp"
    actual = JobPortMapping.from_docker_dict(d[container_port][0], container_port=container_port)
    assert actual.host_port == 32769
    assert actual.container_port == 8888
    assert actual.host_ip == "0.0.0.0"


def test_port_mapping_from_docker_dict_b():
    d = {"HostIp": "127.0.0.1", "HostPort": "1111/tcp", "ContainerPort": "12345/udp"}
    actual = JobPortMapping.from_docker_dict(d)
    assert actual.host_port == 1111
    assert actual.container_port == 12345
    assert actual.host_ip == "127.0.0.1"

def test_matches_gpu_model_1():
    reqResources = JobRequestResources(cpus=1, gpus=1, cpu_memory_gb=1,
                                       gpu_model=["1080", "blah"])
    assert reqResources.matches_gpu_model("nvidia GeForce 1080")
    assert reqResources.matches_gpu_model("1080 GeForce")
    assert reqResources.matches_gpu_model("a b1080c dd")
    assert reqResources.matches_gpu_model("1080")
    assert not reqResources.matches_gpu_model("108")
    assert reqResources.matches_gpu_model("blah")
    assert reqResources.matches_gpu_model("foo blah bar")
    assert reqResources.matches_gpu_model("BLAH")
    assert reqResources.matches_gpu_model("fooBLAHbar")
    assert reqResources.matches_gpu_model("bLAh")
    assert reqResources.matches_gpu_model("bLaHbLaHbLaH")


def test_matches_gpu_model_2():
    reqResources = JobRequestResources(cpus=1, gpus=1, cpu_memory_gb=1,
                                       gpu_model=["blah"])
    assert not reqResources.matches_gpu_model("nvidia GeForce 1080")
    assert not reqResources.matches_gpu_model("1080 GeForce")
    assert not reqResources.matches_gpu_model("a b1080c dd")
    assert not reqResources.matches_gpu_model("1080")
    assert not reqResources.matches_gpu_model("108")
    assert reqResources.matches_gpu_model("blah")
    assert reqResources.matches_gpu_model("foo blah bar")
    assert reqResources.matches_gpu_model("BLAH")
    assert reqResources.matches_gpu_model("fooBLAHbar")
    assert reqResources.matches_gpu_model("bLAh")
    assert reqResources.matches_gpu_model("bLaHbLaHbLaH")


def test_matches_gpu_model_3():
    reqResources = JobRequestResources(cpus=1, gpus=1, cpu_memory_gb=1,
                                       gpu_model=[])
    assert not reqResources.matches_gpu_model("nvidia GeForce 1080")
    assert not reqResources.matches_gpu_model("1080 GeForce")
    assert not reqResources.matches_gpu_model("a b1080c dd")
    assert not reqResources.matches_gpu_model("1080")
    assert not reqResources.matches_gpu_model("108")
    assert not reqResources.matches_gpu_model("blah")
    assert not reqResources.matches_gpu_model("foo blah bar")
    assert not reqResources.matches_gpu_model("BLAH")
    assert not reqResources.matches_gpu_model("fooBLAHbar")
    assert not reqResources.matches_gpu_model("bLAh")
    assert not reqResources.matches_gpu_model("bLaHbLaHbLaH")


def test_matches_gpu_model_4():
    req_resources = lambda req_model: JobRequestResources(cpus=1, gpus=1, cpu_memory_gb=1, gpu_model=[req_model])
    assert req_resources("A100").matches_gpu_model("NVIDIA A100 80GB PCIe (80 GB Mem)")
    assert req_resources("a100").matches_gpu_model("NVIDIA A100 80GB PCIe (80 GB Mem)")
    assert req_resources("A100 80GB").matches_gpu_model("NVIDIA A100 80GB PCIe (80 GB Mem)")
    assert req_resources("a100 80").matches_gpu_model("NVIDIA A100 80GB PCIe (80 GB Mem)")
    assert req_resources("(80 GB Mem)").matches_gpu_model("NVIDIA A100 80GB PCIe (80 GB Mem)")
    assert req_resources("(80 GB MEM)").matches_gpu_model("NVIDIA A100 80GB PCIe (80 GB Mem)")
    assert req_resources("NVIDIA A100 80GB PCIe (80 GB Mem)").matches_gpu_model("NVIDIA A100 80GB PCIe (80 GB Mem)")
    assert req_resources("nvidia a100 80gb pcie (80 gb mem)").matches_gpu_model("NVIDIA A100 80GB PCIe (80 GB Mem)")
    assert not req_resources(" NVIDIA A100 80gb PCIe (80 GB Mem)").matches_gpu_model("NVIDIA A100 80GB PCIe (80 GB Mem)")
    assert not req_resources("nvidia a100 80gb pcie (80 gb mem) ").matches_gpu_model("NVIDIA A100 80GB PCIe (80 GB Mem)")


def test_docker_request_group_add_json1():
    json_in = TEST_JOB2_REQUEST_DOCKER_JSON
    req_dock = JobRequestDocker.from_json(json_in)
    assert req_dock.group_add == ["demogroup"]
    assert req_dock == TEST_JOB2_OBJ_A.request.docker


def test_docker_request_group_add_json2():
    # should also support json/dict with single string instead of list
    json_in = TEST_JOB2_REQUEST_DOCKER_JSON.replace('["demogroup"]', '"demogroup"')
    req_dock = JobRequestDocker.from_json(json_in)
    assert req_dock.group_add == ["demogroup"]
    assert req_dock == TEST_JOB2_OBJ_A.request.docker


def test_docker_request_group_add_dict1():
    dict_in = TEST_JOB2_OBJ_A.request.docker.to_dict()
    dict_in['groupAdd'] = ["demogroup"]
    req_dock = JobRequestDocker.from_dict(dict_in)
    assert req_dock.group_add == ["demogroup"]
    assert req_dock == TEST_JOB2_OBJ_A.request.docker


def test_docker_request_group_add_dict2():
    # should also support json/dict with single string instead of list
    dict_in = TEST_JOB2_OBJ_A.request.docker.to_dict()
    dict_in['groupAdd'] = "demogroup"
    req_dock = JobRequestDocker.from_dict(dict_in)
    assert req_dock.group_add == ["demogroup"]
    assert req_dock == TEST_JOB2_OBJ_A.request.docker


def test_resource_request_gpu_model_dict1():
    dict_in = TEST_JOB2_OBJ_A.request.resources.to_dict()
    dict_in['gpuModel'] = ['TITAN']
    req_resources = JobRequestResources.from_dict(dict_in)
    assert req_resources.gpu_model == ['TITAN']
    assert req_resources == dataclasses.replace(TEST_JOB2_OBJ_A.request.resources, gpu_model=['TITAN'])


def test_resource_request_gpu_model_dict2a():
    dict_in = TEST_JOB2_OBJ_A.request.resources.to_dict()
    dict_in['gpuModel'] = 'TITAN'
    req_resources = JobRequestResources.from_dict(dict_in)
    assert req_resources.gpu_model == ['TITAN']
    assert req_resources == dataclasses.replace(TEST_JOB2_OBJ_A.request.resources, gpu_model=['TITAN'])


def test_resource_request_gpu_model_dict2b():
    dict_in = TEST_JOB2_OBJ_A.to_dict()
    dict_in['request']['resources']['gpuModel'] = 'TITAN'
    job = Job.from_dict(dict_in)
    assert job.request.resources.gpu_model == ['TITAN']
    assert job.request.resources == dataclasses.replace(TEST_JOB2_OBJ_A.request.resources, gpu_model=['TITAN'])


def test_resource_request_gpu_model_dict2c():
    dict_in = TEST_JOB2_OBJ_A.request.to_dict()
    dict_in['resources']['gpuModel'] = 'TITAN'
    job_request = JobRequest.from_dict(dict_in)
    assert job_request.resources.gpu_model == ['TITAN']
    assert job_request.resources == dataclasses.replace(TEST_JOB2_OBJ_A.request.resources, gpu_model=['TITAN'])


PERFORMANCE_TEST_SIZE = 2000  # dataclasses_json implementation of from_dict was 45.782864 s for 10000 !!!
# PERFORMANCE_TEST_MAX_MS_PER = 1.1
PERFORMANCE_TEST_MAX_MS_PER = 5.0

def performance_measure():
    dict_in = TEST_JOB2_OBJ_A.to_dict()
    for i in range(1, PERFORMANCE_TEST_SIZE):
        job_out = Job.from_dict(dict_in)
        assert TEST_JOB2_OBJ_A == job_out


# def test_job2_performance():
#     start = datetime.datetime.now(datetime.timezone.utc)
#     import cProfile
#     with cProfile.Profile() as pr:
#         performance_measure()
#     # pr.print_stats(sort=SortKey.TIME)
#     pr.print_stats(sort=SortKey.CUMULATIVE)
#     end = datetime.datetime.now(datetime.timezone.utc)
#     diff: timedelta = end - start
#     each_ms = (diff.total_seconds()*1000.0)/PERFORMANCE_TEST_SIZE
#     print(f'performance: {diff.total_seconds()} s for {PERFORMANCE_TEST_SIZE} -> {each_ms} ms each')
#     assert diff.total_seconds() < ((PERFORMANCE_TEST_MAX_MS_PER * PERFORMANCE_TEST_SIZE) / 1000.0), \
#         f'bad Job.from_dict() performance: {diff.total_seconds()} s for {PERFORMANCE_TEST_SIZE} ' \
#         f'-> {each_ms} ms each, which is > max {PERFORMANCE_TEST_MAX_MS_PER} ms each'


def test_job_auto_restart_info_a():
    job1 = Job(
        uuid=TEST_JOB2_UUID_A,
        name="testJobA",
        deployment_environment="testing",
        request=TEST_JOB2_REQUEST_OBJ_A,
        description="Test Job A",
        owner=TEST_JOB2_OWNER_OBJ_A,
        state=TEST_JOB2_STATE_OBJ_A,
        restart_info=None,
    )
    job2 = Job(
        uuid=TEST_JOB2_UUID_A,
        name="testJobA",
        deployment_environment="testing",
        request=TEST_JOB2_REQUEST_OBJ_A,
        description="Test Job A",
        owner=TEST_JOB2_OWNER_OBJ_A,
        state=TEST_JOB2_STATE_OBJ_A,
        restart_info=RestartInfo(
            initial_job_uuid=TEST_JOB2_UUID_A,
            restart_count=0,
        ),
    )
    assert job1.restart_info
    assert job1.restart_info.initial_job_uuid == job1.uuid
    assert job1.restart_info.restart_count == 0

    assert job1 == job2

    actual = Job.from_json(TEST_JOB2_JSON)
    assert actual == job1

    assert 'restartInfo' in job1.to_json()
    assert 'restartInfo' in job1.to_dict()


def test_job2_scheduling_reservation_ids_single_a():
    expected = TEST_JOB2_REQUEST_SCHEDULING_OBJ_A
    dict_in = expected.to_dict()
    dict_in['reservationId'] = dict_in['reservationIds'][0]  # only first element
    del dict_in['reservationIds']
    actual = JobRequestScheduling.from_dict(dict_in)
    assert actual == expected


def test_job2_scheduling_reservation_ids_single_b():
    expected = TEST_JOB2_REQUEST_SCHEDULING_OBJ_A
    dict_in = expected.to_dict()
    dict_in['reservationId'] = dict_in['reservationIds']   # entire list
    del dict_in['reservationIds']
    actual = JobRequestScheduling.from_dict(dict_in)
    assert actual == expected


def test_job2_scheduling_reservation_ids_single_c():
    expected = TEST_JOB2_REQUEST_SCHEDULING_OBJ_A
    dict_in = expected.to_dict()
    dict_in['reservationId'] = 'nonsense'
    #del dict_in['reservationIds']  # keep
    actual = JobRequestScheduling.from_dict(dict_in)
    assert actual == expected


def test_job2_scheduling_reservation_ids_single_d():
    expected = TEST_JOB2_REQUEST_SCHEDULING_OBJ_A
    dict_in = expected.to_dict()

    dict_in['reservationId'] = None  # explicit None
    del dict_in['reservationIds']

    expected = dataclasses.replace(expected, reservation_ids=[])

    actual = JobRequestScheduling.from_dict(dict_in)
    assert actual == expected


def test_job2_scheduling_reservation_ids_single_e():
    expected = TEST_JOB2_REQUEST_SCHEDULING_OBJ_A
    dict_in = expected.to_dict()

    dict_in['reservationId'] = None  # explicit None
    dict_in['reservationIds'] = None

    expected = dataclasses.replace(expected, reservation_ids=[])

    actual = JobRequestScheduling.from_dict(dict_in)
    assert actual == expected


def test_job2_scheduling_reservation_ids_single_f():
    expected = TEST_JOB2_REQUEST_SCHEDULING_OBJ_A
    dict_in = expected.to_dict()

    dict_in['reservationIds'] = None

    expected = dataclasses.replace(expected, reservation_ids=[])

    actual = JobRequestScheduling.from_dict(dict_in)
    assert actual == expected


def test_job2_scheduling_reservation_ids_single_g():
    expected = TEST_JOB2_REQUEST_SCHEDULING_OBJ_A
    dict_in = expected.to_dict()

    dict_in['reservationId'] = []
    del dict_in['reservationIds']

    expected = dataclasses.replace(expected, reservation_ids=[])

    actual = JobRequestScheduling.from_dict(dict_in)
    assert actual == expected


def test_job2_scheduling_reservation_ids_single_h():
    expected = TEST_JOB2_REQUEST_SCHEDULING_OBJ_A
    dict_in = expected.to_dict()

    dict_in['reservationId'] = []
    dict_in['reservationIds'] = []

    expected = dataclasses.replace(expected, reservation_ids=[])

    actual = JobRequestScheduling.from_dict(dict_in)
    assert actual == expected


def test_job2_scheduling_reservation_ids_single_i():
    expected = TEST_JOB2_REQUEST_SCHEDULING_OBJ_A
    dict_in = expected.to_dict()

    dict_in['reservationIds'] = []

    expected = dataclasses.replace(expected, reservation_ids=[])

    actual = JobRequestScheduling.from_dict(dict_in)
    assert actual == expected
